import celery
from api_dnl import settings
from datetime import datetime,timedelta
from pytz import UTC
from time import mktime
import io,csv,gzip,zipfile
import os,shutil
import xlwt
import json
from api_dnl.utils.imp_exp import (csv2xls,dict_to_csv)
from celery import Celery
from celery.app.log import get_logger
from celery.schedules import crontab
from falcon_rest.db import initialize_db
from sqlalchemy import (
    Integer, SmallInteger, Float, Text, String, DateTime, Date, Time, Boolean, ForeignKey, BigInteger,
    Table
)
from sqlalchemy import (Column, desc, and_,or_,not_, text as text_, PrimaryKeyConstraint, inspect, Index, UniqueConstraint)
from sqlalchemy.sql import func, select,alias,case
from api_dnl.base_model import DnlApiBaseModel
from api_dnl.model import Client,Client,Resource,ResourcePrefix,Rate,RateTable,RateSendLog,BalanceHistoryActual,C4ClientBalance, \
    RateSendLogDetail,CrdReportDetailTable,MailSender,FraudDetection,FraudDetectionLog,FraudDetectionLogDetail,\
    ResourceBlock,IngressTrunk,SystemParameter,AlertRule,AlertRuleLog,AlertRulesLogDetail,CdrReportDetail,client_cdr,\
    SendRateDirect,MailTemplate,SendRateTemplate,ProductRoutRateTable,ProductClientsRef
from api_dnl.model import PrimaryKeyConstraint,func,query_to_sting,cast,generate_uuid_str,aliased
from api_dnl.tasks import app,log,db,SqlAlchemyTask,SchLog,custom_logger
from api_dnl import model
from api_dnl.utils.statisticapi2 import StatisticAPI
from api_dnl.scheme import RateScheme

@app.task(base=SqlAlchemyTask)
def do_send_rate():
    log.debug('Send rate started')
    sr=RateSendLog.filter(RateSendLog.status=='waiting').all()
    log.debug('Send rate waiting {}'.format(len(sr)))
    for r in RateSendLog.filter(RateSendLog.status=='waiting').all():
        send_rate.delay(r.id)

@app.task(base=SqlAlchemyTask, time_limit=288000, soft_time_limit=287400)
def send_rate(rate_id):
    vlog = custom_logger('send_rate')
    slog = SchLog('send_rate')
    vlog.debug('Send rate log id={} sheduled.'.format(rate_id))
    sr = RateSendLog.get(rate_id)
    if not sr:
        vlog.debug('Send rate log id={} sheduled. but not found'.format(rate_id))
        return False
    # if sr.status != 'waiting':
    #     vlog.warning('Send rate log id={} sheduled but not waiting! skip.'.format(rate_id))
    #     return False
    if not sr.email_template and not sr.email_direct:
        sr.status = 'sent failed'
        sr.error = 'No email template and no direct email!'
        sr.save()
        vlog.warning('Send rate log id={} No email template and no direct email!'.format(rate_id))
        return False
    try:
        if not sr.download_deadline:
            sr.download_deadline = (datetime.now(UTC) + timedelta(days=2)).date()
        sr.status = 'generated'
        sr.error = 'temporary busy '
        sr.save()
        vlog.debug('Send rate log id={}'.format(rate_id))
        rt = RateTable.get(sr.rate_table_id)
        if not rt:
            sr.status = 'sent failed'
            sr.error = 'No rate_table_id '
            sr.save()
            vlog.warning('Send rate log id={} no rate_table_id'.format(rate_id))
            return False

        if sr.start_effective_date:
            objects_list = Rate.filter(and_(Rate.rate_table_id == sr.rate_table_id,
                                            Rate.effective_date <= sr.start_effective_date,
                                            or_(Rate.end_date.is_(None),
                                                Rate.end_date >= sr.start_effective_date)))
        else:
            objects_list = Rate.filter(and_(Rate.rate_table_id == sr.rate_table_id,
                                            #Rate.effective_date <= func.now(),
                                            or_(Rate.end_date.is_(None), Rate.end_date >= func.now())))

        if not sr.effective_date:
            if sr.start_effective_date:
                qed = Rate.session().query(Rate.effective_date).filter(and_(Rate.rate_table_id == sr.rate_table_id,
                                            Rate.effective_date <= sr.start_effective_date,
                                            or_(Rate.end_date.is_(None),
                                                Rate.end_date >= sr.start_effective_date))).group_by(
                    Rate.effective_date).all()
            else:
                qed=Rate.session().query(Rate.effective_date).filter(and_(Rate.rate_table_id == sr.rate_table_id,
                             # Rate.effective_date <= func.now(),
                             or_(Rate.end_date.is_(None), Rate.end_date >= func.now()))).group_by(Rate.effective_date).all()
            if len(qed)==1:
                eff_date = str(qed[0].effective_date)
            else:
                eff_date = 'Varying Dates'
        else:
            eff_date = str(sr.effective_date)



        hdrs=[]
        if sr.email_template:
            hdrs = sr.email_template.headers
            tpl = sr.email_template
        else:
            hdrs = sr.email_direct.headers
            tpl = sr.email_direct
        if sr.indicate_inc_dec:
            hdrs += ['change_status']
        class scheme_class(RateScheme):
            class Meta:
                model = Rate
                fields = hdrs if sr.rate_table.jur_type != 'US Jurisdictional' else hdrs[:hdrs.index('rate')] + ['indeterminte'] + hdrs[hdrs.index('rate') + 1:]
        sr.headers = ','.join(hdrs if sr.rate_table.jur_type != 'US Jurisdictional' else hdrs[:hdrs.index('rate')] + ['indeterminte'] + hdrs[hdrs.index('rate') + 1:])

        vlog.debug('Send rate log data ready')


        att=None
        if sr.format in ['csv','xls']:
            if not objects_list.first():
                sr.status = 'sent failed'
                sr.err = 'No rates for rate_table_id {}'.format(sr.rate_table_id)
                sr.save()
                vlog.warning('Send rate log id={} rate_table_id={} empty '.format(rate_id, sr.rate_table_id))
                return
            csvfile = io.StringIO()
            #fieldnames = data[0].keys()
            writer = csv.DictWriter(csvfile, fieldnames=hdrs if sr.rate_table.jur_type != 'US Jurisdictional' else hdrs[:hdrs.index('rate')] + ['indeterminte'] + hdrs[hdrs.index('rate') + 1:])
            writer.writeheader()
            i = 0
            window_size = 1000  # or whatever limit you like
            window_idx = 0
            while True:
                start, stop = window_size * window_idx, window_size * (window_idx + 1)
                things = objects_list.slice(start, stop).all()
                if things is None:
                    break
                for rate in things:
                    setattr(rate, 'indeterminte', rate.rate)
                    data = scheme_class().dump(rate).data
                    if sr.date_format == 'Date Only':
                        if 'effective_date' in data and data['effective_date']:
                            data['effective_date'] = data['effective_date'][0:10]
                        if 'end_date' in data and data['end_date']:
                            data['end_date'] = data['end_date'][0:10]

                    if sr.effective_date and 'effective_date' in hdrs:
                        eff_d = sr.effective_date[0:10]  # str(sr.effective_date.date())
                        data['effective_date'] = eff_d
                    writer.writerow(data)
                    i += 1
                vlog.debug('Send rate {} records done'.format(i))
                if len(things) < window_size:
                    vlog.debug('len(things) < window_size')
                    break
                window_idx += 1
            sr.total_records = i
            vlog.debug('Send rate log total_records={}'.format(sr.total_records))
            acsv=csvfile.getvalue()
            csvfile.seek(0)
            vlog.debug('Send rate log csv file ready')
            sr.completed_records = sum(1 for line in csvfile)-1
            att = [('Rate_send_{}'.format(generate_uuid_str()()) + '.' + 'csv',acsv)]
            if sr.format=='xls':
                att = [('xls',csv2xls(acsv,'Rate'))]
        elif  sr.format=='json':
            data = scheme_class().dump(objects_list.all(), many=True).data
            sr.completed_records = len(data)
            att = [('txt', json.dumps(data))]
        if sr.zip and att:
            out = io.BytesIO()
            name = 'rate.' + att[0][0]
            with zipfile.ZipFile(out, mode = "w") as f:
                f.writestr(name,att[0][1],zipfile.ZIP_DEFLATED)
            att = [('zip', out.getvalue())]
            vlog.debug('Send rate log zip file ready')
        method='attachment'
        #if sr.download_method == 'link' or
        if  ( sr.email_direct and sr.email_direct.download_method == 'send as link') or \
                                    (sr.email_template and sr.email_template.download_method == 'send as link'):
            method='link'
        if  True: #method=='link':
            sr.file = 'Rate_send_{}.{}'.format(generate_uuid_str()(),att[0][0])
            file_path = os.path.expanduser(settings.FILES['upload_to']) + '/' + sr.file
            try:
                f=open(file_path, 'wb')
                f.write(att[0][1])
            except TypeError:
                f = open(file_path, 'wt')
                f.write(att[0][1])
            att=[]
            f.close()
        sr.status = 'generated'

        sr.save()
        vlog.warning('Send rate log id={} rate_table_id={} attachment generated '.format(rate_id, sr.rate_table_id))
        errors=0
        err_list=[]

        if sr.send_type in ('Send to carriers using this rate table','Send to specified trunks'):
            if sr.send_type == 'Send to carriers using this rate table':
                if sr.trunks==[]:
                    res_pref=aliased(ResourcePrefix)
                    q = Resource.query().outerjoin(res_pref).filter(or_(Resource.rate_table_id == sr.rate_table_id,
                                    res_pref.rate_table_id == sr.rate_table_id)).group_by(Resource.resource_id).all()
                    trunks = []
                    for tr in q:
                        trunks.append(tr.resource_id)
                    sr.trunks = trunks
                else:
                    sr.send_type = 'Send to specified trunks'
            for resource_id in sr.trunks:
                if resource_id=='':
                    continue
                trunk = Resource.get(resource_id)
                if trunk and trunk.client:
                    try:
                        email = getattr(trunk.client,tpl.to_mail,'')
                        if email is None or email == '':
                            email=trunk.client.email
                        l = RateSendLogDetail(log_id=rate_id, resource_id=resource_id, send_to=email, status='In Progress')
                        l.save()
                        trunk.download_deadline = str(sr.download_deadline)
                        if method == 'link':
                            trunk.download_link = str(l.download_link)
                            trunk.download_token = str(l.download_token)
                        else:
                            trunk.download_link = 'download in attachment link not applicable'
                            trunk.download_token = 'download in attachment link not applicable'
                        trunk.effective_date = eff_date
                        trunk.fmt_billing_type = rt.rate_type
                        trunk.fmt_jur_type = rt.jur_type
                        if sr.email_template:
                            res=trunk.apply_mail('sendrate_'+str(sr.email_template_id), sr.send_specify_email, att=att)
                        else:
                            res = trunk.apply_mail('sendratedirect_' + str(sr.id), sr.send_specify_email, att=att)
                        vlog.debug('Send rate apply mail res {}'.format(res))
                        if res:
                            if hasattr(MailSender,'last_error'):
                                mail_sender_last_error=MailSender.last_error
                            else:
                                mail_sender_last_error = 'unknown mail send error'
                            l.error = mail_sender_last_error
                            err_list.append(mail_sender_last_error)
                            l.status='failed'
                            errors +=1
                        else:
                            if method == 'link':
                                l.status = 'Not Yet Downloaded'
                            else:
                                l.status = 'completed'
                            l.sent_on = datetime.now(UTC)
                            sr.save()
                    except Exception as e:
                        l.error=str(e)[0:500]
                        l.status='failed'
                    l.save()
        else:
            class Rec:
                pass
            emails=sr.send_specify_email.split(';')
            for email in emails:
                try:
                    if not email or email == '':
                        continue
                    trunk=Rec()
                    trunk.email = email
                    trunk.rate_email = email
                    trunk.client_name = email
                    trunk.effective_date = eff_date
                    l = RateSendLogDetail(log_id=rate_id, resource_id=None, send_to=email,
                                          status='In Progress')
                    l.save()
                    trunk.download_deadline = str(sr.download_deadline)
                    if method == 'link':
                        trunk.download_link = str(l.download_link)
                        trunk.download_token = str(l.download_token)
                    else:
                        trunk.download_link = 'download in attachment link not applicable'
                        trunk.download_token = 'download in attachment link not applicable'
                    trunk.fmt_billing_type = rt.rate_type
                    trunk.fmt_jur_type = rt.jur_type
                    trunk.fmt_rate_table_name = rt.name
                    trunk.fmt_allowed_ip = 'n/a'
                    trunk.fmt_switch_ip = rt.switch_ip
                    if sr.email_template:
                        res = MailSender.apply_mail(trunk,'sendrate_' + str(sr.email_template_id), email, att=att)
                    else:
                        res = MailSender.apply_mail(trunk,'sendratedirect_' + str(sr.id), email, att=att)
                    vlog.debug('Send rate apply mail res {}'.format(res))
                    if res:
                        if hasattr(MailSender, 'last_error'):
                            mail_sender_last_error = MailSender.last_error
                        else:
                            mail_sender_last_error = 'unknown mail send error'
                        l.error = mail_sender_last_error
                        err_list.append(mail_sender_last_error)
                        l.status = 'failed'
                        errors += 1
                    else:
                        if method == 'link':
                            l.status = 'Not Yet Downloaded'
                        else:
                            l.status = 'completed'
                        l.sent_on = datetime.now(UTC)
                        sr.save()
                except Exception as e:
                    l.error = str(e)[0:500]
                    l.status = 'failed'
                    l.save()


        if not errors:
            sr.status='sent success'
            sr.error = None
            sr.save()
            vlog.debug('Send rate finished with success')
        else:
            try:
                err_list = set(err_list)
                e=','.join(err_list)
            except:
                e='mail send error'
            vlog.debug('Send rate finished with srrors'.format(e))
            sr.status = 'sent failed'
            sr.err = 'Send rate mail failed {} times'.format(e)
            sr.save()
            vlog.warning('Send rate id={} mail errors={}'.format(rate_id, e))
    except Exception as e:
        sr.status = 'sent failed'
        sr.err = 'Send rate id={} error={}'.format(rate_id, str(e))
        try:
            sr.save()
            vlog.error('Send rate id={} error={}'.format(rate_id, str(e)))
        except Exception as e:
            vlog.error('Send rate id={} fatal error={}'.format(rate_id, str(e)))
            db.session.rollback()
        return False
    slog.close()
    return True



@app.task(base=SqlAlchemyTask,time_limit=1800,soft_time_limit=1790)
def send_rate_task(rate_table_id,clients):
    slog = SchLog('send_rate_task')
    sr = RateSendLog(rate_table_id=rate_table_id, status='waiting')
    try:
        sr.error = 'temporary busy'
        tpl = MailTemplate.get('rate')
        email_direct = SendRateDirect(sender_id=tpl.from_mail_id,subject=tpl.subject,content=tpl.html_content,
                                      #header=SendRateTemplate.HEADERS,
                                      download_method = 'send as link',
                                      mail_cc=tpl.cc_mail)
        sr.email_direct = email_direct

        trunks = []
        for client_id in clients:
            cl = Client.get(client_id)
            if cl.resources and len(cl.resources):
                trunks.append(cl.resources[0].resource_id)
        sr.trunks = trunks
        sr.send_type == 'Send to specified trunks'
        sr.format = 'csv'
        rate_id = sr.save()
        return send_rate(rate_id)
    except Exception as e:
        sr.status = 'sent failed'
        sr.err = 'Send rate task error={}'.format( str(e))
        try:
            sr.save()
            log.error('Send rate task error={}'.format( str(e)))
        except Exception as e:
            log.error('Send rate task fatal error={}'.format( str(e)))
            db.session.rollback()
        return False
    slog.close()
    return True


@app.task(base=SqlAlchemyTask,time_limit=1800,soft_time_limit=1790)
def send_client_product_to_emails_task(client_id,product_id,emails):
    job_id = None
    slog = SchLog('send_rate_to_emails_task')
    pr = ProductRoutRateTable.get(product_id)
    sr = RateSendLog(rate_table_id=pr.rate_table_id, status='waiting')
    try:
        sr.error = 'temporary busy'
        tpl = MailTemplate.get('rate')
        email_direct = SendRateDirect(sender_id=tpl.from_mail_id,subject=tpl.subject,content=tpl.html_content,
                                      #header=SendRateTemplate.HEADERS,
                                      download_method='send as link',
                                      mail_cc=tpl.cc_mail)
        sr.email_direct = email_direct

        sr.send_type == 'Specify my own recipients'
        sr.send_specify_email = emails
        sr.format = 'csv'
        job_id = sr.save()
        if job_id:#send_rate(job_id):
            q=ProductClientsRef.filter(and_(ProductClientsRef.product_id==product_id,ProductClientsRef.client_id==client_id))
            if not q:
                q=ProductClientsRef(product_id==product_id,client_id=client_id)
            q.rate_last_sent=datetime.now(UTC)
            q.save()
        return job_id

    except Exception as e:
        sr.status = 'sent failed'
        sr.err = 'Send rate task error={}'.format( str(e))
        try:
            job_id = sr.save()
            log.error('Send rate task error={}'.format( str(e)))
        except Exception as e:
            log.error('Send rate task fatal error={}'.format( str(e)))
            db.session.rollback()
        return job_id
    slog.close()
    return job_id