from .did_api import DidAPI
from api_dnl.model import log, DidBillingRel, Client, Resource
from sqlalchemy import and_, or_, func


class DidAPIManual(DidAPI):
    def __init__(self, user, password=None, vendor=None):
        self.vendor = vendor

    def order_local(self, items, client_id=None):
        log.debug('manual vendor order {}'.format(items))
        cls = DidBillingRel
        ret = []
        for i in items:
            did = cls.filter(
                and_(or_(cls.did.like('_' + i['number']), cls.did == i['number']), cls.is_toll_free == False,
                     cls.end_date.is_(None),
                     cls.egress_res_id.is_(None), cls.ingress_res_id.isnot(None),
                     or_(cls.assigned_client_id.is_(None),
                         cls.assigned_client_id == int(client_id)))).first()
            if did:
                vendor_id = did.vendor_res.client_id if did.vendor_res else None
                ret.append(dict(number=i['number'], vendor_id=vendor_id, status='success'))
            else:
                ret.append(
                    dict(number=i['number'], status='error', error='number:{} not in repository'.format(i['number'])))
        log.debug('manual vendor returned:{}'.format(ret))
        return ret

    def search_local(self, state=None, npa=None, lata=None, pattern=None, count=100, client_id=None):
        log.debug('manual vendor search state={} lata={} pattern={} count={}'.format(state, lata, pattern, count))
        cls = DidBillingRel
        filt = and_(cls.is_toll_free == False, cls.end_date.is_(None), cls.egress_res_id.is_(None),
                    cls.ingress_res_id.isnot(None), cls.is_available,
                    or_(cls.assigned_client_id.is_(None), cls.assigned_client_id == int(client_id)))
        if npa is not None:
            filt = and_(filt, or_(
                and_(func.substr(cls.did, 2, 3) == npa, func.length(cls.did) == 11),
                and_(func.substr(cls.did, 1, 3) == npa, func.length(cls.did) < 10),
            )
                        )
        if pattern is not None:
            pat = pattern.replace('*', '_').replace('x', '_') + '%'
            pat1 = '_' + pattern
            filt = and_(filt, or_(cls.did.like(pat), cls.did.like(pat1)))
        if state is not None:
            filt = and_(filt, cls.state == state)
        if lata is not None:
            filt = and_(filt, cls.lata == lata)
        ret = []
        for did in cls.filter(filt):
            vendor_id = did.vendor_res.client_id if did.vendor_res else None
            ret.append(dict(number=did.did, vendor_id=vendor_id, country=did.country,state=did.state,lata=did.lata,ocn=did.ocn))
        log.debug('manual vendor returned:{}'.format(ret))
        return ret

    def search_toll_free(self, pattern=None, count=100, client_id=None):
        log.debug('manual vendor search pattern={} count={}'.format(pattern, count))
        cls = DidBillingRel
        filt = and_(cls.is_toll_free, cls.end_date.is_(None), cls.egress_res_id.is_(None),
                    cls.ingress_res_id.isnot(None), cls.is_available,
                    or_(cls.assigned_client_id.is_(None), cls.assigned_client_id == int(client_id)))
        if pattern is not None:
            pat = pattern.replace('*', '_').replace('x', '_') + '%'
            pat1 = '_' + pattern
            filt = and_(filt, or_(cls.did.like(pat), cls.did.like(pat1)))
        ret = []
        for did in cls.filter(filt):
            vendor_id = did.vendor_res.client_id if did.vendor_res else None
            ret.append(dict(number=did.did, vendor_id=vendor_id, country=did.country,state=did.state,lata=did.lata,ocn=did.ocn))
        log.debug('manual vendor returned:{}'.format(ret))
        return ret

    def order_toll_free(self, items, routing_label=None, client_id=None):
        log.debug('manual vendor order {}'.format(items))
        cls = DidBillingRel
        ret = []
        for i in items:
            did = cls.filter(
                and_(or_(cls.did.like('_' + i['number']), cls.did == i['number']), cls.is_toll_free == True,
                     cls.end_date.is_(None),
                     cls.egress_res_id.is_(None), cls.ingress_res_id.isnot(None),
                     or_(cls.assigned_client_id.is_(None),
                         cls.assigned_client_id == int(client_id)))).first()
            log.debug('manual vendor did found {}'.format(did))
            if did:
                vendor_id = did.vendor_res.client_id if did.vendor_res else None
                ret.append(dict(number=i['number'], vendor_id=vendor_id, status='sucess', hint='manual'))
            else:
                ret.append(
                    dict(number=i['number'], status='error', error='number:{} not in repository'.format(i['number'])))
        log.debug('manual vendor returned:{}'.format(ret))
        return ret

    def pending(self):
        return []


class VendorManual(object):

    def __init__(self):
        self.vendor = self
        self.id = None
        self.name = 'Local manual repository'

    def get_api(self):
        return DidAPIManual(None, None, self)

    pass
